# frozen_string_literal: true

require 'delegate'
require_relative 'local_variable/version'

class Ractor
  class LocalVariable
    module Implementation
      MUTEX = Ractor.new do
        loop do
          Ractor.yield :obtained
          Ractor.receive
        end
      end
      def MUTEX.synchronize
        take # => :obtained
        begin
          yield
        ensure
          send :release
        end
      end
      private_constant :MUTEX

      def initialize(&block)
        raise ArgumentError, 'No block given' unless block

        @initializer = Ractor.make_shareable(block)
        @hold_me = Object.new.freeze
        @unique_id = :"Ractor::LocalVariable<#{@hold_me.object_id}>"
        freeze
      end

      def ractor_local_value
        t = Thread.main
        t.thread_variable_get(@unique_id) ||
          MUTEX.synchronize do
            t.thread_variable_get(@unique_id) ||
              t.thread_variable_set(@unique_id, @initializer.call ||
                raise(NotImplementedError,
                      'Initializer blocks returning `nil` of `false` not supported'))
          end
      end
    end

    include Implementation

    def self.new(delegate: false, &block)
      return Delegator.new(&block) if delegate

      super()
    end

    class Delegator < ::Delegator
      include Implementation
      alias_method :__getobj__, :ractor_local_value
      undef_method :marshal_load, :initialize_clone, :initialize_dup
      define_method :freeze, Kernel.instance_method(:freeze) # do not delegate `freeze`
    end
  end
end
